# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import List, cast

from rp2.balance import BalanceSet
from rp2.configuration import Configuration
from rp2.entry_types import EntrySetType, TransactionType
from rp2.gain_loss_set import GainLossSet
from rp2.in_transaction import InTransaction
from rp2.input_data import InputData
from rp2.rp2_decimal import ZERO, RP2Decimal
from rp2.rp2_error import RP2TypeError, RP2ValueError
from rp2.transaction_set import TransactionSet


@dataclass(frozen=True, eq=True)
class YearlyGainLoss:
    year: int
    asset: str
    transaction_type: TransactionType
    is_long_term_capital_gains: bool
    crypto_amount: RP2Decimal
    usd_amount: RP2Decimal
    usd_cost_basis: RP2Decimal
    usd_gain_loss: RP2Decimal

    def __post_init__(self) -> None:
        Configuration.type_check_positive_int("year", self.year)
        Configuration.type_check_string("asset", self.asset)
        TransactionType.type_check("transaction_type", self.transaction_type)
        Configuration.type_check_bool("is_long_term", self.is_long_term_capital_gains)
        Configuration.type_check_decimal("crypto_amount", self.crypto_amount)
        Configuration.type_check_decimal("usd_amount", self.usd_amount)
        Configuration.type_check_decimal("usd_cost_basis", self.usd_cost_basis)
        Configuration.type_check_decimal("usd_gain_loss", self.usd_gain_loss)

    def __eq__(self, other: object) -> bool:
        if not other:
            return False
        if not isinstance(other, YearlyGainLoss):
            raise RP2TypeError(f"Operand has non-YearlyGainLoss value {repr(other)}")
        result: bool = (
            self.year == other.year
            and self.asset == other.asset
            and self.transaction_type == other.transaction_type
            and self.is_long_term_capital_gains == other.is_long_term_capital_gains
        )
        return result

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __lt__(self, other: object) -> bool:
        if not other:
            return False
        if not isinstance(other, YearlyGainLoss):
            raise RP2TypeError(f"Operand has non-YearlyGainLoss value {repr(other)}")
        return self.year < other.year

    def __gt__(self, other: object) -> bool:
        if not other:
            return False
        if not isinstance(other, YearlyGainLoss):
            raise RP2TypeError(f"Operand has non-YearlyGainLoss value {repr(other)}")
        return self.year > other.year

    def __ge__(self, other: object) -> bool:
        return not self.__lt__(other)

    def __le__(self, other: object) -> bool:
        return not self.__gt__(other)

    def __hash__(self) -> int:
        return hash((self.year, self.asset, self.transaction_type, self.is_long_term_capital_gains))


class ComputedData:
    @classmethod
    def type_check(cls, name: str, instance: "ComputedData") -> "ComputedData":
        Configuration.type_check_parameter_name(name)
        if not isinstance(instance, cls):
            raise RP2TypeError(f"Parameter '{name}' is not of type {cls.__name__}: {instance}")
        return instance

    @staticmethod
    def _filter_yearly_gain_loss_by_year(yearly_gain_loss_list: List[YearlyGainLoss], from_year: int, to_year: int) -> List[YearlyGainLoss]:
        return [y for y in yearly_gain_loss_list if y.year >= from_year and y.year <= to_year]

    # from_year is not used when computing average price per unit (because we always start from the beginning): only to_year is relevant.
    @staticmethod
    def _compute_price_per_unit(in_transaction_set: TransactionSet, to_year: int) -> RP2Decimal:
        crypto_in_running_sum: RP2Decimal = ZERO
        usd_in_with_fee_running_sum: RP2Decimal = ZERO
        for entry in in_transaction_set:
            if entry.timestamp.year > to_year:
                break
            transaction: InTransaction = cast(InTransaction, entry)
            crypto_in_running_sum += transaction.crypto_in
            usd_in_with_fee_running_sum += transaction.usd_in_with_fee
        return usd_in_with_fee_running_sum / crypto_in_running_sum if crypto_in_running_sum is not ZERO else ZERO

    def __init__(
        self,
        asset: str,
        taxable_event_set: TransactionSet,
        gain_loss_set: GainLossSet,
        yearly_gain_loss_list: List[YearlyGainLoss],
        input_data: InputData,
        from_year: int,
        to_year: int,
    ) -> None:
        InputData.type_check("input_data", input_data)
        Configuration.type_check_positive_int("from_year", from_year)
        Configuration.type_check_positive_int("to_year", to_year, non_zero=True)

        self.__asset: str = Configuration.type_check_string("asset", asset)
        self.__taxable_event_set: TransactionSet = TransactionSet.type_check("taxable_event_set", taxable_event_set, EntrySetType.MIXED, asset, True)
        self.__gain_loss_set: GainLossSet = GainLossSet.type_check("gain_loss_set", gain_loss_set)

        if not isinstance(yearly_gain_loss_list, List):
            raise RP2TypeError(f"Parameter 'yearly_gain_loss_list' is not of type List: {yearly_gain_loss_list}")
        self.__yearly_gain_loss_list: List[YearlyGainLoss] = self._filter_yearly_gain_loss_by_year(yearly_gain_loss_list, from_year, to_year)

        self.__in_transaction_set: TransactionSet = input_data.in_transaction_set
        self.__intra_transaction_set: TransactionSet = input_data.intra_transaction_set
        self.__out_transaction_set: TransactionSet = input_data.out_transaction_set

        self.__balance_set: BalanceSet = BalanceSet(taxable_event_set.configuration, input_data, to_year)
        self.__price_per_unit: RP2Decimal = self._compute_price_per_unit(input_data.in_transaction_set, to_year)

        if self.__taxable_event_set.asset != self.__asset:
            raise RP2ValueError(f"Asset mismatch in 'taxable_event_set': expected {self.__asset}, found {self.__taxable_event_set.asset}")
        if self.__gain_loss_set.asset != self.__asset:
            raise RP2ValueError(f"Asset mismatch in 'gain_loss_set': expected {self.__asset}, found {self.__gain_loss_set.asset}")
        if self.__balance_set.asset != self.__asset:
            raise RP2ValueError(f"Asset mismatch in 'balance_set': expected {self.__asset}, found {self.__balance_set.asset}")

        if self.__asset != input_data.asset:
            raise RP2ValueError(f"Asset mismatch in 'input_data': expected {self.__asset}, found {input_data.asset}")

    @property
    def asset(self) -> str:
        return self.__asset

    @property
    def taxable_event_set(self) -> TransactionSet:
        return self.__taxable_event_set

    @property
    def gain_loss_set(self) -> GainLossSet:
        return self.__gain_loss_set

    @property
    def yearly_gain_loss_list(self) -> List[YearlyGainLoss]:
        return self.__yearly_gain_loss_list

    @property
    def in_transaction_set(self) -> TransactionSet:
        return self.__in_transaction_set

    @property
    def out_transaction_set(self) -> TransactionSet:
        return self.__out_transaction_set

    @property
    def intra_transaction_set(self) -> TransactionSet:
        return self.__intra_transaction_set

    @property
    def balance_set(self) -> BalanceSet:
        return self.__balance_set

    @property
    def price_per_unit(self) -> RP2Decimal:
        return self.__price_per_unit



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from argparse import ArgumentParser, Namespace
from difflib import unified_diff
from itertools import zip_longest
from pathlib import Path
from typing import Any, List

import ezodf
from rp2.rp2_decimal import CRYPTO_DECIMALS


def _row_as_string(row: Any) -> str:
    values: List[str] = []
    cell: Any
    for cell in row:
        value: Any = str(_parse_cell_value(cell))
        values.append(value)

    # Remove trailing whitespace from row
    i: int
    for i in range(len(values) - 1, -1, -1):
        if not values[i]:
            del values[i]
        else:
            break

    return ",".join(values)


def _parse_cell_value(cell: Any) -> Any:
    value: Any
    try:
        if cell.value or cell.value == 0:
            value = round(float(cell.value), CRYPTO_DECIMALS)
            if value == -0.0:
                value = 0
        else:
            value = ""
    except ValueError:
        value = cell.value

    return value


def ods_diff(file1_path: Path, file2_path: Path) -> str:

    if not file1_path.exists():
        return f"Error: {file1_path} does not exist"
    if not file2_path.exists():
        return f"Error: {file2_path} does not exist"

    file1: Any = ezodf.opendoc(str(file1_path))
    file2: Any = ezodf.opendoc(str(file2_path))
    sheet1: Any
    sheet2: Any

    contents1: List[str] = []
    contents2: List[str] = []
    for sheet1 in file1.sheets:
        if sheet1.name not in file2.sheets.names():
            contents2.append(f"{sheet1.name}: sheet not found in '{file2_path}'")
            continue
        sheet2 = file2.sheets[sheet1.name]
        contents1.append(sheet1.name)
        contents2.append(sheet1.name)

        row1: Any = None
        row2: Any = None
        for _, (row1, row2) in enumerate(zip_longest(sheet1.rows(), sheet2.rows())):
            if row1:
                contents1.append(_row_as_string(row1))
            if row2:
                contents2.append(_row_as_string(row2))

    for sheet2 in file2.sheets:
        if sheet2.name not in file1.sheets.names():
            contents1.append(f"{sheet2.name}: sheet not found in '{file1_path}'")

    return "\n".join(unified_diff(contents1, contents2, lineterm=""))


def main() -> None:

    parser: ArgumentParser = ArgumentParser(description="Generate yearly capital gain/loss report and account balances for crypto holdings.")
    parser.add_argument(
        "ods_file1",
        action="store",
        help="First ODS file",
        metavar="ODS_FILE1",
        type=str,
    )
    parser.add_argument(
        "ods_file2",
        action="store",
        help="Second ODS file",
        metavar="ODS_FILE2",
        type=str,
    )

    args: Namespace = parser.parse_args()
    print(ods_diff(Path(args.ods_file1), Path(args.ods_file2)))


if __name__ == "__main__":
    main()



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import unittest
from pathlib import Path
from subprocess import run

from ods_diff import ods_diff

ROOT_PATH: Path = Path(os.path.dirname(__file__)).parent.absolute()

CONFIG_PATH: Path = ROOT_PATH / Path("config")
INPUT_PATH: Path = ROOT_PATH / Path("input")
GOLDEN_PATH: Path = INPUT_PATH / Path("golden")
LOG_PATH: Path = ROOT_PATH / Path("log")
OUTPUT_PATH: Path = ROOT_PATH / Path("output")


class TestODSOutputDiff(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # Generate output to compare with golden files
        shutil.rmtree(LOG_PATH, ignore_errors=True)
        shutil.rmtree(OUTPUT_PATH, ignore_errors=True)
        run(
            ["rp2", "-o", str(OUTPUT_PATH), "-p", "test_data_", str(CONFIG_PATH / Path("test_data.config")), str(INPUT_PATH / Path("test_data.ods"))],
            check=True,
        )
        run(
            [
                "rp2",
                "-o",
                str(OUTPUT_PATH),
                "-p",
                "test_many_year_data_",
                str(CONFIG_PATH / Path("test_data.config")),
                str(INPUT_PATH / Path("test_many_year_data.ods")),
            ],
            check=True,
        )
        run(
            [
                "rp2",
                "-o",
                str(OUTPUT_PATH),
                "-p",
                "crypto_example_",
                str(CONFIG_PATH / Path("crypto_example.config")),
                str(INPUT_PATH / Path("crypto_example.ods")),
            ],
            check=True,
        )

    def setUp(self) -> None:
        self.maxDiff = None  # pylint: disable=invalid-name

    def test_data_rp2_full_report_plugin(self) -> None:
        diff: str = ods_diff(
            GOLDEN_PATH / Path("test_data_rp2_full_report.ods"),
            OUTPUT_PATH / Path("test_data_rp2_full_report.ods"),
        )
        self.assertFalse(diff, msg=diff)

    def test_data_tax_report_us_plugin(self) -> None:
        diff: str = ods_diff(
            GOLDEN_PATH / Path("test_data_tax_report_us.ods"),
            OUTPUT_PATH / Path("test_data_tax_report_us.ods"),
        )
        self.assertFalse(diff, msg=diff)

    def test_many_year_data_rp2_full_report_plugin(self) -> None:
        diff: str = ods_diff(
            GOLDEN_PATH / Path("test_many_year_data_rp2_full_report.ods"),
            OUTPUT_PATH / Path("test_many_year_data_rp2_full_report.ods"),
        )
        self.assertFalse(diff, msg=diff)

    def test_many_year_data_tax_report_us_plugin(self) -> None:
        diff: str = ods_diff(
            GOLDEN_PATH / Path("test_many_year_data_tax_report_us.ods"),
            OUTPUT_PATH / Path("test_many_year_data_tax_report_us.ods"),
        )
        self.assertFalse(diff, msg=diff)

    def test_crypto_example_rp2_full_report_plugin(self) -> None:
        diff: str = ods_diff(
            GOLDEN_PATH / Path("crypto_example_rp2_full_report.ods"),
            OUTPUT_PATH / Path("crypto_example_rp2_full_report.ods"),
        )
        self.assertFalse(diff, msg=diff)

    def test_crypto_example_tax_report_us_plugin(self) -> None:
        diff: str = ods_diff(
            GOLDEN_PATH / Path("crypto_example_tax_report_us.ods"),
            OUTPUT_PATH / Path("crypto_example_tax_report_us.ods"),
        )
        self.assertFalse(diff, msg=diff)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Dict, Iterable, Iterator, List, Set, cast

from rp2.abstract_entry import AbstractEntry
from rp2.abstract_transaction import AbstractTransaction
from rp2.computed_data import ComputedData, YearlyGainLoss
from rp2.configuration import Configuration
from rp2.entry_types import TransactionType
from rp2.gain_loss import GainLoss
from rp2.gain_loss_set import GainLossSet
from rp2.in_transaction import InTransaction
from rp2.input_data import InputData
from rp2.intra_transaction import IntraTransaction
from rp2.logger import LOGGER
from rp2.out_transaction import OutTransaction
from rp2.rp2_decimal import ZERO, RP2Decimal
from rp2.rp2_error import RP2ValueError
from rp2.transaction_set import TransactionSet


def compute_tax(configuration: Configuration, input_data: InputData) -> ComputedData:
    Configuration.type_check("configuration", configuration)
    InputData.type_check("input_data", input_data)

    taxable_event_set: TransactionSet = _create_taxable_event_set(configuration, input_data)
    LOGGER.debug("%s: Created taxable event set", input_data.asset)
    gain_loss_set: GainLossSet = _create_gain_and_loss_set(configuration, input_data, taxable_event_set)
    LOGGER.debug("%s: Created gain-loss set", input_data.asset)
    yearly_gain_loss_list: List[YearlyGainLoss] = _create_yearly_gain_loss_list(input_data, gain_loss_set)
    LOGGER.debug("%s: Created yearly gain-loss list", input_data.asset)

    return ComputedData(
        input_data.asset,
        taxable_event_set,
        gain_loss_set,
        yearly_gain_loss_list,
        input_data,
        configuration.from_year,
        configuration.to_year,
    )


def _create_taxable_event_set(configuration: Configuration, input_data: InputData) -> TransactionSet:
    transaction_set: TransactionSet
    entry: AbstractEntry
    transaction: AbstractTransaction
    taxable_event_set: TransactionSet = TransactionSet(configuration, "MIXED", input_data.asset, configuration.from_year, configuration.to_year)
    for transaction_set in [
        input_data.in_transaction_set,
        input_data.out_transaction_set,
        input_data.intra_transaction_set,
    ]:
        for entry in transaction_set:
            transaction = cast(AbstractTransaction, entry)
            if transaction.is_taxable():
                taxable_event_set.add_entry(transaction)

    return taxable_event_set


def _create_gain_and_loss_set(configuration: Configuration, input_data: InputData, taxable_event_set: TransactionSet) -> GainLossSet:

    gain_loss_set: GainLossSet = GainLossSet(configuration, input_data.asset, from_year=configuration.from_year, to_year=configuration.to_year)

    taxable_event_iterator: Iterator[AbstractTransaction] = iter(cast(Iterable[AbstractTransaction], taxable_event_set))
    from_lot_iterator: Iterator[InTransaction] = iter(cast(Iterable[InTransaction], input_data.in_transaction_set))

    try:
        gain_loss: GainLoss
        taxable_event: AbstractTransaction = next(taxable_event_iterator)
        taxable_event_amount: RP2Decimal = taxable_event.crypto_taxable_amount
        from_lot: InTransaction = next(from_lot_iterator)
        from_lot_amount: RP2Decimal = from_lot.crypto_in

        while True:
            if taxable_event.transaction_type.is_earn_type():
                # Handle earn-typed transactions first
                gain_loss = GainLoss(configuration, taxable_event_amount, taxable_event, None)
                gain_loss_set.add_entry(gain_loss)
                taxable_event = next(taxable_event_iterator)
                taxable_event_amount = taxable_event.crypto_taxable_amount
                continue

            if taxable_event_amount == from_lot_amount:
                gain_loss = GainLoss(configuration, taxable_event_amount, taxable_event, from_lot)
                gain_loss_set.add_entry(gain_loss)
                taxable_event = next(taxable_event_iterator)
                taxable_event_amount = taxable_event.crypto_taxable_amount
                from_lot = next(from_lot_iterator)
                from_lot_amount = from_lot.crypto_in
            elif taxable_event_amount < from_lot_amount:
                gain_loss = GainLoss(configuration, taxable_event_amount, taxable_event, from_lot)
                gain_loss_set.add_entry(gain_loss)
                from_lot_amount = from_lot_amount - taxable_event_amount
                taxable_event = next(taxable_event_iterator)
                taxable_event_amount = taxable_event.crypto_taxable_amount
            else:  # taxable_amount > from_lot_amount
                gain_loss = GainLoss(configuration, from_lot_amount, taxable_event, from_lot)
                gain_loss_set.add_entry(gain_loss)
                taxable_event_amount = taxable_event_amount - from_lot_amount
                from_lot = next(from_lot_iterator)
                from_lot_amount = from_lot.crypto_in
    except StopIteration as exception:
        if cast(Iterator[InTransaction], exception.value) == from_lot_iterator:
            raise RP2ValueError("Total in-transaction value < total taxable entries") from None

    return gain_loss_set


@dataclass(frozen=True, eq=True)
class _YearlyGainLossId:
    year: int
    asset: str
    transaction_type: TransactionType
    is_long_term_capital_gains: bool


# Frozen and eq are not set because we don't need to hash instances and we need to modify fields (see _create_yearly_gain_loss_list)
@dataclass(frozen=True, eq=True)
class _YearlyGainLossAmounts:
    crypto_amount: RP2Decimal
    usd_amount: RP2Decimal
    usd_cost_basis: RP2Decimal
    usd_gain_loss: RP2Decimal


def _create_yearly_gain_loss_list(input_data: InputData, gain_loss_set: GainLossSet) -> List[YearlyGainLoss]:
    summaries: Dict[_YearlyGainLossId, _YearlyGainLossAmounts] = {}
    entry: AbstractEntry
    key: _YearlyGainLossId
    value: _YearlyGainLossAmounts
    for entry in gain_loss_set:
        gain_loss: GainLoss = cast(GainLoss, entry)
        key = _YearlyGainLossId(
            gain_loss.taxable_event.timestamp.year,
            gain_loss.asset,
            gain_loss.taxable_event.transaction_type,
            gain_loss.is_long_term_capital_gains(),
        )
        value = summaries.setdefault(key, _YearlyGainLossAmounts(ZERO, ZERO, ZERO, ZERO))
        crypto_amount: RP2Decimal = value.crypto_amount + gain_loss.crypto_amount
        usd_amount: RP2Decimal = value.usd_amount + gain_loss.taxable_event_usd_amount_with_fee_fraction
        usd_cost_basis: RP2Decimal = value.usd_cost_basis + gain_loss.usd_cost_basis
        usd_gain_loss: RP2Decimal = value.usd_gain_loss + gain_loss.usd_gain
        summaries[key] = _YearlyGainLossAmounts(crypto_amount=crypto_amount, usd_amount=usd_amount, usd_cost_basis=usd_cost_basis, usd_gain_loss=usd_gain_loss)

    yearly_gain_loss_set: Set[YearlyGainLoss] = set()
    crypto_taxable_amount_total: RP2Decimal = ZERO
    usd_taxable_amount_total: RP2Decimal = ZERO
    cost_basis_total: RP2Decimal = ZERO
    gain_loss_total: RP2Decimal = ZERO
    for (key, value) in summaries.items():
        yearly_gain_loss: YearlyGainLoss = YearlyGainLoss(
            year=key.year,
            asset=key.asset,
            transaction_type=key.transaction_type,
            is_long_term_capital_gains=key.is_long_term_capital_gains,
            crypto_amount=value.crypto_amount,
            usd_amount=value.usd_amount,
            usd_cost_basis=value.usd_cost_basis,
            usd_gain_loss=value.usd_gain_loss,
        )
        yearly_gain_loss_set.add(yearly_gain_loss)
        crypto_taxable_amount_total += yearly_gain_loss.crypto_amount
        usd_taxable_amount_total += yearly_gain_loss.usd_amount
        cost_basis_total += yearly_gain_loss.usd_cost_basis
        gain_loss_total += yearly_gain_loss.usd_gain_loss

    _verify_computation(input_data, crypto_taxable_amount_total, usd_taxable_amount_total, cost_basis_total, gain_loss_total)

    return list(sorted(yearly_gain_loss_set, key=_yearly_gain_loss_sort_criteria, reverse=True))


def _yearly_gain_loss_sort_criteria(yearly_gain_loss: YearlyGainLoss) -> str:
    return (
        f"{yearly_gain_loss.asset}"
        f" {yearly_gain_loss.year}"
        f" {'LONG' if yearly_gain_loss.is_long_term_capital_gains else 'SHORT'}"
        f" {yearly_gain_loss.transaction_type.value}"
    )


# Internal sanity check: ensure the sum total of gains and losses from taxable flows matches the one from taxable events and InTransactions
def _verify_computation(
    input_data: InputData,
    crypto_taxable_amount_total: RP2Decimal,
    usd_taxable_amount_total: RP2Decimal,
    cost_basis_total: RP2Decimal,
    gain_loss_total: RP2Decimal,
) -> None:
    usd_taxable_amount_total_verify: RP2Decimal = ZERO
    crypto_earned_amount_total_verify: RP2Decimal = ZERO
    crypto_sold_amount_total_verify: RP2Decimal = ZERO
    crypto_taxable_amount_total_verify: RP2Decimal = ZERO
    cost_basis_total_verify: RP2Decimal = ZERO
    crypto_in_amount_total: RP2Decimal = ZERO
    gain_loss_total_verify: RP2Decimal = ZERO

    # Compute USD and crypto total taxable amount
    for entry in input_data.in_transaction_set:
        in_transaction: InTransaction = cast(InTransaction, entry)
        usd_taxable_amount_total_verify += in_transaction.usd_taxable_amount
        crypto_earned_amount_total_verify += in_transaction.crypto_taxable_amount
    for entry in input_data.out_transaction_set:
        out_transaction: OutTransaction = cast(OutTransaction, entry)
        usd_taxable_amount_total_verify += out_transaction.usd_taxable_amount
        crypto_sold_amount_total_verify += out_transaction.crypto_taxable_amount
    for entry in input_data.intra_transaction_set:
        intra_transaction: IntraTransaction = cast(IntraTransaction, entry)
        usd_taxable_amount_total_verify += intra_transaction.usd_taxable_amount
        crypto_sold_amount_total_verify += intra_transaction.crypto_taxable_amount

    crypto_taxable_amount_total_verify = crypto_sold_amount_total_verify + crypto_earned_amount_total_verify

    # Compute cost basis
    for entry in input_data.in_transaction_set:
        in_transaction = cast(InTransaction, entry)
        if crypto_in_amount_total + in_transaction.crypto_in > crypto_sold_amount_total_verify:
            # End of loop: last in transaction covering the amount sold needs to be fractioned
            crypto_in_amount: RP2Decimal = crypto_sold_amount_total_verify - crypto_in_amount_total
            crypto_in_amount_total += crypto_in_amount
            cost_basis_total_verify += (crypto_in_amount / in_transaction.crypto_in) * in_transaction.usd_in_with_fee
            break
        crypto_in_amount_total += in_transaction.crypto_in
        cost_basis_total_verify += in_transaction.usd_in_with_fee

    gain_loss_total_verify = usd_taxable_amount_total_verify - cost_basis_total_verify

    if crypto_taxable_amount_total != crypto_taxable_amount_total_verify:
        raise Exception(
            f"{input_data.asset}: total crypto taxable amount incongruence detected: " f"{crypto_taxable_amount_total} != {crypto_taxable_amount_total_verify}",
        )
    if usd_taxable_amount_total != usd_taxable_amount_total_verify:
        raise Exception(
            f"{input_data.asset}: total usd taxable amount incongruence detected: " f"{usd_taxable_amount_total} != {usd_taxable_amount_total_verify}",
        )
    if cost_basis_total != cost_basis_total_verify:
        raise Exception(f"{input_data.asset}: cost basis incongruence detected: {cost_basis_total} != {cost_basis_total_verify}")
    if gain_loss_total != gain_loss_total_verify:
        raise Exception(f"{input_data.asset}: total gain/loss incongruence detected: {gain_loss_total} != {gain_loss_total_verify}")



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import unittest
from datetime import datetime
from tempfile import NamedTemporaryFile
from typing import Any, Optional

import jsonschema  # type: ignore
from dateutil.tz import tzoffset, tzutc
from rp2.configuration import AccountingMethod, Configuration
from rp2.rp2_decimal import ZERO, RP2Decimal
from rp2.rp2_error import RP2TypeError, RP2ValueError


class TestConfiguration(unittest.TestCase):
    _configuration: Configuration

    @classmethod
    def setUpClass(cls) -> None:
        TestConfiguration._configuration = Configuration("./config/test_data.config")

    def setUp(self) -> None:
        self.maxDiff = None  # pylint: disable=invalid-name

    @staticmethod
    def _test_config(config: Any) -> Configuration:
        result: Optional[Configuration] = None
        with NamedTemporaryFile(delete=False) as temporary_file:
            temporary_file.write(json.dumps(config).encode())
            temporary_file.flush()

            result = Configuration(temporary_file.name)
        os.remove(temporary_file.name)

        return result

    def test_config_file(self) -> None:
        config: Any = {}
        with self.assertRaisesRegex(KeyError, "in_header"):
            self._test_config(config)

        config["in_header"] = None
        with self.assertRaisesRegex(jsonschema.exceptions.ValidationError, "None is not of type 'object'"):
            self._test_config(config)

        config["in_header"] = {}
        with self.assertRaisesRegex(jsonschema.exceptions.ValidationError, "'timestamp' is a required property"):
            self._test_config(config)

        config["in_header"].update(
            {
                "timestamp": 0,
                "asset": 6,
                "exchange": 1,
                "holder": 2,
                "transaction_type": 5,
            }
        )
        with self.assertRaisesRegex(jsonschema.exceptions.ValidationError, "'spot_price' is a required property"):
            self._test_config(config)

        config["in_header"].update({"spot_price": 8, "crypto_in": 7, "usd_fee": 11, "usd_in_no_fee": 9, "usd_in_with_fee": 10, "notes": 12})
        with self.assertRaisesRegex(KeyError, "out_header"):
            self._test_config(config)

        config["out_header"] = "foobar"
        with self.assertRaisesRegex(jsonschema.exceptions.ValidationError, "'foobar' is not of type 'object'"):
            self._test_config(config)

        config["out_header"] = {}
        config["out_header"].update({"timestamp": 0, "asset": 6, "exchange": 1, "holder": 2, "transaction_type": 5})
        with self.assertRaisesRegex(jsonschema.exceptions.ValidationError, "'spot_price' is a required property"):
            self._test_config(config)

        config["out_header"].update({"spot_price": 8, "crypto_out_no_fee": 7, "crypto_fee": 9, "notes": 12})
        with self.assertRaisesRegex(KeyError, "intra_header"):
            self._test_config(config)

        config["intra_header"] = [1, 2, 3]
        with self.assertRaisesRegex(jsonschema.exceptions.ValidationError, r"\[1, 2, 3\] is not of type 'object'"):
            self._test_config(config)

        config["intra_header"] = {}
        config["intra_header"].update(
            {
                "timestamp": 0,
                "asset": 6,
                "from_exchange": 1,
                "from_holder": 2,
                "to_exchange": 3,
            }
        )
        with self.assertRaisesRegex(jsonschema.exceptions.ValidationError, "'to_holder' is a required property"):
            self._test_config(config)

        config["intra_header"].update({"to_holder": 4, "spot_price": 8, "crypto_sent": 7, "crypto_received": 10, "notes": 12})
        with self.assertRaisesRegex(KeyError, "assets"):
            self._test_config(config)

        config["assets"] = "foobar"
        with self.assertRaisesRegex(jsonschema.exceptions.ValidationError, "'foobar' is not of type 'array'"):
            self._test_config(config)

        config["assets"] = []
        with self.assertRaisesRegex(jsonschema.exceptions.ValidationError, r"\[\] is too short"):
            self._test_config(config)

        config["assets"] = [1, 2, 3]
        with self.assertRaisesRegex(jsonschema.exceptions.ValidationError, "1 is not of type 'string'"):
            self._test_config(config)

        config["assets"] = ["B1", "B2"]
        with self.assertRaisesRegex(KeyError, "exchanges"):
            self._test_config(config)

        config["exchanges"] = None
        with self.assertRaisesRegex(jsonschema.exceptions.ValidationError, "None is not of type 'array'"):
            self._test_config(config)

        config["exchanges"] = []
        with self.assertRaisesRegex(jsonschema.exceptions.ValidationError, r"\[\] is too short"):
            self._test_config(config)

        config["exchanges"] = [1, 2, 3]
        with self.assertRaisesRegex(jsonschema.exceptions.ValidationError, "1 is not of type 'string'"):
            self._test_config(config)

        config["exchanges"] = ["BlockFi", "Coinbase", "Kraken"]
        with self.assertRaisesRegex(KeyError, "holders"):
            self._test_config(config)

        config["holders"] = 7
        with self.assertRaisesRegex(jsonschema.exceptions.ValidationError, "7 is not of type 'array'"):
            self._test_config(config)

        config["holders"] = []
        with self.assertRaisesRegex(jsonschema.exceptions.ValidationError, r"\[\] is too short"):
            self._test_config(config)

        config["holders"] = [1, 2, 3]
        with self.assertRaisesRegex(jsonschema.exceptions.ValidationError, "1 is not of type 'string'"):
            self._test_config(config)

        config["holders"] = ["Bob"]
        self.assertIsNotNone(self._test_config(config))

        config["in_header"].update({"spot_price": -8})
        with self.assertRaisesRegex(jsonschema.exceptions.ValidationError, "-8 is less than the minimum of 0"):
            self._test_config(config)

        config["in_header"].update({"spot_price": 8})
        config["out_header"].update({"notes": -12})
        with self.assertRaisesRegex(jsonschema.exceptions.ValidationError, "-12 is less than the minimum of 0"):
            self._test_config(config)

        config["out_header"].update({"notes": 12})
        config["intra_header"].update({"crypto_sent": -7})
        with self.assertRaisesRegex(jsonschema.exceptions.ValidationError, "-7 is less than the minimum of 0"):
            self._test_config(config)

        config["intra_header"].update({"crypto_sent": 7})
        config["assets"].append("B1")
        with self.assertRaisesRegex(jsonschema.exceptions.ValidationError, ".* has non-unique elements"):
            self._test_config(config)

        config["assets"] = ["B1", "B2"]

        config["in_header"].update({"foobar": 23})
        with self.assertRaisesRegex(jsonschema.exceptions.ValidationError, "Additional properties are not allowed .*'foobar' was unexpected.*"):
            self._test_config(config)

    def test_creation(self) -> None:
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'configuration_path' has non-string value .*"):
            Configuration(None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'configuration_path' has non-string value .*"):
            Configuration(111)  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "/non/existing/file does not exist"):
            Configuration("/non/existing/file")
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'accounting_method' is not of type AccountingMethod: .*"):
            Configuration("./config/test_data.config", accounting_method=None)  # type: ignore
        with self.assertRaisesRegex(NotImplementedError, ".*'AccountingMethod.LIFO' not implemented yet"):
            Configuration("./config/test_data.config", accounting_method=AccountingMethod.LIFO)
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'from_year' has non-integer value .*"):
            Configuration("./config/test_data.config", from_year="foobar")  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'from_year' has non-positive value .*"):
            Configuration("./config/test_data.config", from_year=-1)
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'to_year' has non-integer value .*"):
            Configuration("./config/test_data.config", to_year="foobar")  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'to_year' has non-positive value .*"):
            Configuration("./config/test_data.config", to_year=-1)
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'to_year' has zero value"):
            Configuration("./config/test_data.config", to_year=0)

    def test_argument_packs(self) -> None:
        self.assertEqual(
            str(self._configuration.get_in_table_constructor_argument_pack([0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100, 110, 120])),
            (
                "{'timestamp': 0, 'asset': 60, 'exchange': 10, 'holder': 20, 'transaction_type': 50, 'spot_price': 80, "
                "'crypto_in': 70, 'usd_fee': 110, 'usd_in_no_fee': 90, 'usd_in_with_fee': 100, 'notes': 120}"
            ),
        )
        self.assertEqual(
            str(self._configuration.get_out_table_constructor_argument_pack([0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100, 110, 120])),
            (
                "{'timestamp': 0, 'asset': 60, 'exchange': 10, 'holder': 20, 'transaction_type': 50, 'spot_price': 80, "
                "'crypto_out_no_fee': 70, 'crypto_fee': 90, 'notes': 120}"
            ),
        )
        self.assertEqual(
            str(self._configuration.get_intra_table_constructor_argument_pack([0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100, 110, 120])),
            (
                "{'timestamp': 0, 'asset': 60, 'from_exchange': 10, 'from_holder': 20, 'to_exchange': 30, 'to_holder': 40, "
                "'spot_price': 80, 'crypto_sent': 70, 'crypto_received': 100, 'notes': 120}"
            ),
        )

        with self.assertRaisesRegex(RP2TypeError, "Parameter 'data' value is not a List"):
            self._configuration.get_in_table_constructor_argument_pack(None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'data' value is not a List"):
            self._configuration.get_out_table_constructor_argument_pack("foobar")  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'data' value is not a List"):
            self._configuration.get_intra_table_constructor_argument_pack(8)  # type: ignore

        with self.assertRaisesRegex(RP2ValueError, "Parameter 'data' has length .*, but required minimum from in-table headers in .* is .*"):
            self._configuration.get_in_table_constructor_argument_pack([1, 2, 3])
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'data' has length .*, but required minimum from out-table headers in .* is .*"):
            self._configuration.get_out_table_constructor_argument_pack([])
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'data' has length .*, but required minimum from intra-table headers in .* is .*"):
            self._configuration.get_intra_table_constructor_argument_pack([1])

    def test_get_column_position(self) -> None:
        self.assertEqual(self._configuration.get_in_table_column_position("timestamp"), 0)
        self.assertEqual(self._configuration.get_in_table_column_position("asset"), 6)
        self.assertEqual(self._configuration.get_in_table_column_position("exchange"), 1)
        self.assertEqual(self._configuration.get_in_table_column_position("holder"), 2)
        self.assertEqual(self._configuration.get_in_table_column_position("transaction_type"), 5)
        self.assertEqual(self._configuration.get_in_table_column_position("spot_price"), 8)
        self.assertEqual(self._configuration.get_in_table_column_position("crypto_in"), 7)
        self.assertEqual(self._configuration.get_in_table_column_position("usd_fee"), 11)
        self.assertEqual(self._configuration.get_in_table_column_position("usd_in_no_fee"), 9)
        self.assertEqual(self._configuration.get_in_table_column_position("usd_in_with_fee"), 10)
        self.assertEqual(self._configuration.get_in_table_column_position("notes"), 12)

        self.assertEqual(self._configuration.get_out_table_column_position("timestamp"), 0)
        self.assertEqual(self._configuration.get_out_table_column_position("asset"), 6)
        self.assertEqual(self._configuration.get_out_table_column_position("exchange"), 1)
        self.assertEqual(self._configuration.get_out_table_column_position("holder"), 2)
        self.assertEqual(self._configuration.get_out_table_column_position("transaction_type"), 5)
        self.assertEqual(self._configuration.get_out_table_column_position("spot_price"), 8)
        self.assertEqual(self._configuration.get_out_table_column_position("crypto_out_no_fee"), 7)
        self.assertEqual(self._configuration.get_out_table_column_position("crypto_fee"), 9)
        self.assertEqual(self._configuration.get_out_table_column_position("notes"), 12)

        self.assertEqual(self._configuration.get_intra_table_column_position("timestamp"), 0)
        self.assertEqual(self._configuration.get_intra_table_column_position("asset"), 6)
        self.assertEqual(self._configuration.get_intra_table_column_position("from_exchange"), 1)
        self.assertEqual(self._configuration.get_intra_table_column_position("from_holder"), 2)
        self.assertEqual(self._configuration.get_intra_table_column_position("to_exchange"), 3)
        self.assertEqual(self._configuration.get_intra_table_column_position("to_holder"), 4)
        self.assertEqual(self._configuration.get_intra_table_column_position("spot_price"), 8)
        self.assertEqual(self._configuration.get_intra_table_column_position("crypto_sent"), 7)
        self.assertEqual(self._configuration.get_intra_table_column_position("crypto_received"), 10)
        self.assertEqual(self._configuration.get_intra_table_column_position("notes"), 12)

        with self.assertRaisesRegex(RP2TypeError, "Parameter 'input_parameter' has non-string value .*"):
            self._configuration.get_in_table_column_position(None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'input_parameter' has non-string value .*"):
            self._configuration.get_out_table_column_position(12)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'input_parameter' has non-string value .*"):
            self._configuration.get_intra_table_column_position([1, 2])  # type: ignore

        with self.assertRaisesRegex(RP2ValueError, "Unknown 'input_parameter' .*"):
            self._configuration.get_in_table_column_position("non_existent")
        with self.assertRaisesRegex(RP2ValueError, "Unknown 'input_parameter' .*"):
            self._configuration.get_out_table_column_position("foobar")
        with self.assertRaisesRegex(RP2ValueError, "Unknown 'input_parameter' .*"):
            self._configuration.get_intra_table_column_position("whatever")

    def test_string_representation(self) -> None:
        self.assertEqual(
            str(self._configuration),
            (
                "Configuration(configuration_path=./config/test_data.config, accounting_method=AccountingMethod.FIFO, from_year=non-specified, "
                "to_year=non-specified, in_header={'timestamp': 0, 'asset': 6, 'exchange': 1, 'holder': 2, 'transaction_type': 5, 'spot_price': 8, "
                "'crypto_in': 7, 'usd_fee': 11, 'usd_in_no_fee': 9, 'usd_in_with_fee': 10, 'notes': 12}, out_header={'timestamp': 0, 'asset': 6, "
                "'exchange': 1, 'holder': 2, 'transaction_type': 5, 'spot_price': 8, 'crypto_out_no_fee': 7, 'crypto_fee': 9, 'notes': 12}, "
                "intra_header={'timestamp': 0, 'asset': 6, 'from_exchange': 1, 'from_holder': 2, 'to_exchange': 3, 'to_holder': 4, "
                "'spot_price': 8, 'crypto_sent': 7, 'crypto_received': 10, 'notes': 12}, assets=['B1', 'B2', 'B3', 'B4'], "
                "exchanges=['BlockFi', 'Coinbase', 'Coinbase Pro', 'Kraken'], holders=['Alice', 'Bob'])"
            ),
        )

    def test_unique_id(self) -> None:
        self.assertEqual(13, self._configuration.type_check_unique_id("unique_id", 13))

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            self._configuration.type_check_unique_id(None, 10)  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'unique_id' has non-positive value .*"):
            self._configuration.type_check_unique_id("unique_id", -23)
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'unique_id' has non-integer value .*"):
            self._configuration.type_check_unique_id("unique_id", 7.7)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'unique_id' has non-integer value .*"):
            self._configuration.type_check_unique_id("unique_id", None)  # type: ignore

    def test_timestamp(self) -> None:
        date: datetime = self._configuration.type_check_timestamp_from_string("timestamp", "2020-06-21T23:29:03.117Z")
        self.assertEqual(2020, date.year)
        self.assertEqual(6, date.month)
        self.assertEqual(21, date.day)
        self.assertEqual(23, date.hour)
        self.assertEqual(29, date.minute)
        self.assertEqual(3, date.second)
        self.assertEqual(117000, date.microsecond)
        self.assertEqual(tzutc(), date.tzinfo)

        date = self._configuration.type_check_timestamp_from_string("timestamp", "2020-12-01 03:59:49 -04:00")
        self.assertEqual(2020, date.year)
        self.assertEqual(12, date.month)
        self.assertEqual(1, date.day)
        self.assertEqual(3, date.hour)
        self.assertEqual(59, date.minute)
        self.assertEqual(49, date.second)
        self.assertEqual(0, date.microsecond)
        self.assertEqual(tzoffset(None, -14400), date.tzinfo)

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            self._configuration.type_check_timestamp_from_string(12, "2020-12-01 03:59:49 -04:00")  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'timestamp' has non-string value .*"):
            self._configuration.type_check_timestamp_from_string("timestamp", 2020)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'timestamp' has non-string value .*"):
            self._configuration.type_check_timestamp_from_string("timestamp", None)  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'timestamp' value has no timezone info: .*"):
            self._configuration.type_check_timestamp_from_string("timestamp", "2020-12-01 03:59:49")
        with self.assertRaisesRegex(RP2ValueError, "hour must be in 0..23"):
            self._configuration.type_check_timestamp_from_string("timestamp", "2020-12-01 25:59:49")
        with self.assertRaisesRegex(RP2ValueError, "Unknown string format: .*"):
            self._configuration.type_check_timestamp_from_string("timestamp", "foo bar baz")

    def test_exchange(self) -> None:
        self.assertEqual("Coinbase Pro", self._configuration.type_check_exchange("exchange", "Coinbase Pro"))

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            self._configuration.type_check_exchange(None, "Coinbase")  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'exchange' has non-string value .*"):
            self._configuration.type_check_exchange("exchange", 34.6)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'exchange' has non-string value .*"):
            self._configuration.type_check_exchange("exchange", None)  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'exchange' value is not known: .*"):
            self._configuration.type_check_exchange("exchange", "coinbase pro")
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'exchange' value is not known: .*"):
            self._configuration.type_check_exchange("exchange", "Coinbase Proo")

    def test_holder(self) -> None:
        self.assertEqual("Bob", self._configuration.type_check_holder("holder", "Bob"))

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            self._configuration.type_check_holder(None, "Alice")  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'holder' has non-string value .*"):
            self._configuration.type_check_holder("holder", 34.6)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'holder' has non-string value .*"):
            self._configuration.type_check_holder("holder", None)  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'holder' value is not known: .*"):
            self._configuration.type_check_holder("holder", "John")

    def test_asset(self) -> None:
        self.assertEqual("B1", self._configuration.type_check_asset("asset", "B1"))
        self.assertEqual("B2", self._configuration.type_check_asset("asset", "B2"))

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            self._configuration.type_check_asset(None, "B1")  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'asset' has non-string value .*"):
            self._configuration.type_check_asset("asset", 34.6)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'asset' has non-string value .*"):
            self._configuration.type_check_asset("asset", None)  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'asset' value is not known: .*"):
            self._configuration.type_check_asset("asset", "btc")
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'asset' value is not known: .*"):
            self._configuration.type_check_asset("asset", "qwerty")

    def test_string(self) -> None:
        self.assertEqual("foobar", self._configuration.type_check_string("my_string", "foobar"))

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            self._configuration.type_check_string(None, "foobar")  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_string' has non-string value .*"):
            self._configuration.type_check_string("my_string", [1, 2, 3])  # type: ignore

    def test_int(self) -> None:
        self.assertEqual(5, self._configuration.type_check_int("my_int", 5))
        self.assertEqual(5, self._configuration.type_check_positive_int("my_int", 5))
        self.assertEqual(5, self._configuration.type_check_positive_int("my_int", 5, non_zero=True))

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            self._configuration.type_check_int(None, 5)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_int' has non-integer value .*"):
            self._configuration.type_check_int("my_int", None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_int' has non-integer value .*"):
            self._configuration.type_check_int("my_int", "5")  # type: ignore

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            self._configuration.type_check_positive_int(None, 5)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_int' has non-integer value .*"):
            self._configuration.type_check_positive_int("my_int", None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_int' has non-integer value .*"):
            self._configuration.type_check_positive_int("my_int", "5")  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'my_int' has non-positive value .*"):
            self._configuration.type_check_positive_int("my_int", -5)
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'my_int' has zero value"):
            self._configuration.type_check_positive_int("my_int", 0, non_zero=True)

    def test_float(self) -> None:
        self.assertEqual(5.5, self._configuration.type_check_float("my_float", 5.5))
        self.assertEqual(5.5, self._configuration.type_check_positive_float("my_float", 5.5))
        self.assertEqual(5.5, self._configuration.type_check_positive_float("my_float", 5.5, non_zero=True))

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            self._configuration.type_check_float(None, 5.5)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_float' has non-numeric value .*"):
            self._configuration.type_check_float("my_float", None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_float' has non-numeric value .*"):
            self._configuration.type_check_float("my_float", "5.5")  # type: ignore

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            self._configuration.type_check_positive_float(None, 5.5)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_float' has non-numeric value .*"):
            self._configuration.type_check_positive_float("my_float", None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_float' has non-numeric value .*"):
            self._configuration.type_check_positive_float("my_float", "5.5")  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'my_float' has non-positive value .*"):
            self._configuration.type_check_positive_float("my_float", -5.5)
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'my_float' has zero value"):
            self._configuration.type_check_positive_float("my_float", 0, non_zero=True)

    def test_rp2_decimal(self) -> None:
        minus_one: RP2Decimal = RP2Decimal("-1")
        one: RP2Decimal = RP2Decimal("1")

        self.assertEqual(minus_one, self._configuration.type_check_decimal("minus_one", minus_one))
        self.assertEqual(one, self._configuration.type_check_positive_decimal("one", one))

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            self._configuration.type_check_decimal(None, one)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_decimal' has non-RP2Decimal value .*"):
            self._configuration.type_check_decimal("my_decimal", None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_decimal' has non-RP2Decimal value .*"):
            self._configuration.type_check_decimal("my_decimal", "5.5")  # type: ignore

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            self._configuration.type_check_positive_decimal(None, one)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_decimal' has non-RP2Decimal value .*"):
            self._configuration.type_check_positive_decimal("my_decimal", None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_decimal' has non-RP2Decimal value .*"):
            self._configuration.type_check_positive_decimal("my_decimal", "5.5")  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'my_decimal' has non-positive value .*"):
            self._configuration.type_check_positive_decimal("my_decimal", minus_one)
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'my_decimal' has zero value"):
            self._configuration.type_check_positive_decimal("my_decimal", ZERO, non_zero=True)

    def test_bool(self) -> None:
        self.assertEqual(True, self._configuration.type_check_bool("my_bool", True))
        self.assertEqual(False, self._configuration.type_check_bool("my_bool", False))

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            self._configuration.type_check_bool(14, True)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_bool' has non-bool value .*"):
            self._configuration.type_check_bool("my_bool", None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_bool' has non-bool value .*"):
            self._configuration.type_check_bool("my_bool", "True")  # type: ignore

    def test_accounting_method(self) -> None:
        self.assertEqual(AccountingMethod.FIFO, AccountingMethod.type_check("accounting_method", AccountingMethod.FIFO))

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            AccountingMethod.type_check(None, AccountingMethod.FIFO)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'accounting_method' is not of type AccountingMethod: .*"):
            AccountingMethod.type_check("accounting_method", "fifo")  # type: ignore
        with self.assertRaisesRegex(NotImplementedError, "Parameter 'accounting_method': 'AccountingMethod.LIFO' not implemented yet"):
            AccountingMethod.type_check("accounting_method", AccountingMethod.LIFO)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Any, Dict, List, NamedTuple, Optional, Set, cast

import ezodf
from rp2.abstract_entry import AbstractEntry
from rp2.abstract_transaction import AbstractTransaction
from rp2.balance import BalanceSet
from rp2.computed_data import ComputedData, YearlyGainLoss
from rp2.gain_loss import GainLoss
from rp2.gain_loss_set import GainLossSet
from rp2.in_transaction import InTransaction
from rp2.intra_transaction import IntraTransaction
from rp2.logger import LOGGER
from rp2.out_transaction import OutTransaction
from rp2.plugin.output.abstract_odt_generator import AbstractODTGenerator
from rp2.rp2_decimal import RP2Decimal
from rp2.rp2_error import RP2TypeError
from rp2.transaction_set import TransactionSet


class _TransactionVisualStyle(NamedTuple):
    year: int
    visual_style: str
    highlighted_style: str


class _BorderStyle(NamedTuple):
    year: int
    border_suffix: str


_IN_HEADER_NAMES_ROW_1: List[str] = [
    "",
    "",
    "",
    "",
    "",
    "Transaction",
    "",
    "Crypto",
    "Crypto In",
    "",
    "USD In",
    "USD In",
    "Taxable",
    "",
    "",
]

_IN_HEADER_NAMES_ROW_2: List[str] = [
    "Sent/Sold",
    "Timestamp",
    "Asset",
    "Exchange",
    "Holder",
    "Type",
    "Spot Price",
    "In",
    "Running Sum",
    "USD Fee",
    "No Fee",
    "With Fee",
    "Event",
    "N/A",
    "Notes",
]

_OUT_HEADER_NAMES_ROW_1: List[str] = [
    "",
    "",
    "",
    "",
    "Transaction",
    "",
    "",
    "",
    "Crypto Out",
    "Crypto Fee",
    "",
    "",
    "Taxable",
    "",
]

_OUT_HEADER_NAMES_ROW_2: List[str] = [
    "Timestamp",
    "Asset",
    "Exchange",
    "Holder",
    "Type",
    "Spot Price",
    "Crypto Out",
    "Crypto Fee",
    "Running Sum",
    "Running Sum",
    "USD Out",
    "USD Fee",
    "Event",
    "Notes",
]

_INTRA_HEADER_NAMES_ROW_1: List[str] = [
    "",
    "",
    "From",
    "From",
    "",
    "",
    "",
    "",
    "Crypto",
    "",
    "Crypto Fee",
    "",
    "Taxable",
    "",
]

_INTRA_HEADER_NAMES_ROW_2: List[str] = [
    "Timestamp",
    "Asset",
    "Exchange",
    "Holder",
    "To Exchange",
    "To Holder",
    "Spot Price",
    "Crypto Sent",
    "Received",
    "Crypto Fee",
    "Running Sum",
    "USD Fee",
    "Event",
    "Notes",
]

_BALANCE_HEADER_NAMES_ROW_1: List[str] = [
    "",
    "",
    "",
    "Acquired",
    "Sent",
    "Received",
    "Final",
]

_BALANCE_HEADER_NAMES_ROW_2: List[str] = [
    "Exchange",
    "Holder",
    "Asset",
    "Balance",
    "Balance",
    "Balance",
    "Balance",
]

_GAIN_LOSS_SUMMARY_HEADER_NAMES_ROW_1: List[str] = [
    "",
    "",
    "Capital",
    "Capital",
    "Transaction",
    "Crypto",
    "USD",
    "USD Total",
]

_GAIN_LOSS_SUMMARY_HEADER_NAMES_ROW_2: List[str] = [
    "Year",
    "Asset",
    "Gains",
    "Gains Type",
    "Type",
    "Taxable Total",
    "Taxable Total",
    "Cost Basis",
]

_GAIN_LOSS_DETAIL_HEADER_NAMES_ROW_1: List[str] = [
    "Crypto",
    "",
    "Crypto Amt",
    "Capital",
    "Capital",
    "Taxable Event",
    "Taxable Event",
    "Taxable Event",
    "Taxable Event USD",
    "Taxable Event",
    "Taxable Event",
    "In Lot",
    "In Lot",
    "In Lot USD",
    "In Lot USD",
    "In Lot USD",
    "In Lot",
    "In Lot Fraction",
]

_GAIN_LOSS_DETAIL_HEADER_NAMES_ROW_2: List[str] = [
    "Amount",
    "Asset",
    "Running Sum",
    "Gains",
    "Gains Type",
    "Timestamp",
    "Direction/Type",
    "Fraction %",
    "Amount Fraction",
    "Spot Price",
    "Fraction Description",
    "Timestamp",
    "Fraction %",
    "Amount Fraction",
    "Fee Fraction",
    "Cost Basis",
    "Spot Price",
    "Description",
]

_ZERO: RP2Decimal = RP2Decimal(0)


class Generator(AbstractODTGenerator):

    MIN_ROWS: int = 40
    MAX_COLUMNS: int = 40
    OUTPUT_FILE: str = "rp2_full_report.ods"
    TEMPLATE_SHEETS_TO_KEEP: Set[str] = {"__Legend", "__Summary"}

    def generate(
        self,
        asset_to_computed_data: Dict[str, ComputedData],
        output_dir_path: str,
        output_file_prefix: str,
    ) -> None:

        if not isinstance(asset_to_computed_data, Dict):
            raise RP2TypeError(f"Parameter 'asset_to_computed_data' has non-Dict value {asset_to_computed_data}")

        output_file: Any
        output_file = self._initialize_output_file(
            output_dir_path=output_dir_path,
            output_file_prefix=output_file_prefix,
            output_file_name=self.OUTPUT_FILE,
            template_sheets_to_keep=self.TEMPLATE_SHEETS_TO_KEEP,
        )

        asset: str
        computed_data: ComputedData

        summary_row_index: int = 3
        for asset, computed_data in asset_to_computed_data.items():
            if not isinstance(asset, str):
                raise RP2TypeError(f"Parameter 'asset' has non-string value {asset}")
            ComputedData.type_check("computed_data", computed_data)
            summary_row_index = self.__generate_asset(computed_data, output_file, summary_row_index)

        output_file.save()
        LOGGER.info("Plugin '%s' output: %s", __name__, Path(output_file.docname).resolve())

    def __get_number_of_rows_in_transaction_sheet(self, computed_data: ComputedData) -> int:
        return self.MIN_ROWS + computed_data.in_transaction_set.count + computed_data.out_transaction_set.count + computed_data.intra_transaction_set.count

    def __get_number_of_rows_in_output_sheet(self, computed_data: ComputedData) -> int:
        return self.MIN_ROWS + len(computed_data.yearly_gain_loss_list) + computed_data.balance_set.count + computed_data.gain_loss_set.count

    def __generate_asset(self, computed_data: ComputedData, output_file: Any, summary_row_index: int) -> int:
        asset: str = computed_data.asset
        transaction_sheet_name: str = f"{asset} In-Out"
        output_sheet_name: str = f"{asset} Tax"

        transaction_sheet: Any = ezodf.Table(transaction_sheet_name)
        output_sheet: Any = ezodf.Table(output_sheet_name)
        summary_sheet: Any = output_file.sheets["Summary"]

        output_file.sheets += transaction_sheet
        output_file.sheets += output_sheet

        transaction_sheet.reset(size=(self.__get_number_of_rows_in_transaction_sheet(computed_data), self.MAX_COLUMNS))
        output_sheet.reset(size=(self.__get_number_of_rows_in_output_sheet(computed_data), self.MAX_COLUMNS))

        new_lines: int = len(computed_data.yearly_gain_loss_list)
        if new_lines:
            summary_sheet.append_rows(new_lines)

        row_index: int = 0
        row_index = self.__generate_in_table(transaction_sheet, computed_data.in_transaction_set, computed_data.gain_loss_set, row_index)
        row_index = self.__generate_out_table(transaction_sheet, computed_data.out_transaction_set, row_index + 2)
        row_index = self.__generate_intra_table(transaction_sheet, computed_data.intra_transaction_set, row_index + 2)

        row_index = 0
        row_index = self.__generate_gain_loss_summary(output_sheet, computed_data.yearly_gain_loss_list, row_index)
        row_index = self.__generate_account_balances(output_sheet, computed_data.balance_set, row_index + 2)
        row_index = self.__generate_average_price_per_unit(output_sheet, asset, computed_data.price_per_unit, row_index + 2)
        row_index = self.__generate_gain_loss_detail(output_sheet, asset, computed_data.gain_loss_set, row_index + 2)

        return self.__generate_yearly_gain_loss_summary(summary_sheet, asset, computed_data.yearly_gain_loss_list, summary_row_index)

    @staticmethod
    def __get_transaction_visual_style(transaction: AbstractTransaction, year: int) -> _TransactionVisualStyle:
        visual_style: str = "transparent"
        highlighted_style: str = "transparent"
        if transaction.is_taxable():
            visual_style = "taxable_event"
            highlighted_style = "highlighted"
        if year == 0:
            year = transaction.timestamp.year
        if transaction.timestamp.year != year:
            visual_style = f"{visual_style}_border"
            highlighted_style = f"{highlighted_style}_border"
            year = transaction.timestamp.year
        return _TransactionVisualStyle(year, visual_style, highlighted_style)

    @staticmethod
    def __get_border_style(current_year: int, year: int) -> _BorderStyle:
        border_suffix: str = ""
        if year == 0:
            year = current_year
        if current_year != year:
            border_suffix = "_border"
            year = current_year
        return _BorderStyle(year, border_suffix)

    def __generate_in_table(self, sheet: Any, in_transaction_set: TransactionSet, gain_loss_set: GainLossSet, row_index: int) -> int:
        row_index = self._fill_header("In-Flow Detail", _IN_HEADER_NAMES_ROW_1, _IN_HEADER_NAMES_ROW_2, sheet, row_index, 0)

        in_transaction_index: int = row_index
        entry: AbstractEntry
        crypto_in_running_sum: RP2Decimal = _ZERO
        year: int = 0
        visual_style: str
        for entry in in_transaction_set:
            transaction: InTransaction = cast(InTransaction, entry)
            highlighted_style: str
            transaction_visual_style: _TransactionVisualStyle = self.__get_transaction_visual_style(transaction, year)
            year = transaction_visual_style.year
            visual_style = transaction_visual_style.visual_style
            highlighted_style = transaction_visual_style.highlighted_style
            self._fill_cell(sheet, row_index, 0, "", visual_style="transparent")
            self._fill_cell(sheet, row_index, 1, transaction.timestamp, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 2, transaction.asset, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 3, transaction.exchange, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 4, transaction.holder, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 5, transaction.transaction_type.value.upper(), visual_style=visual_style)
            self._fill_cell(sheet, row_index, 6, transaction.spot_price, data_style="usd", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 7, transaction.crypto_in, data_style="crypto", visual_style=visual_style)
            crypto_in_running_sum += transaction.crypto_in
            self._fill_cell(sheet, row_index, 8, crypto_in_running_sum, data_style="crypto", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 9, transaction.usd_fee, data_style="usd", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 10, transaction.usd_in_no_fee, data_style="usd", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 11, transaction.usd_in_with_fee, data_style="usd", visual_style=highlighted_style)
            self._fill_cell(sheet, row_index, 12, "YES" if transaction.is_taxable() else "NO", data_style="usd", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 13, "", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 14, transaction.notes, visual_style="transparent")
            row_index += 1

        current_from_lot: Optional[AbstractTransaction] = None
        current_from_lot_percentage: RP2Decimal = _ZERO
        year = 0
        border_style: _BorderStyle
        border_suffix: str = ""
        for entry in gain_loss_set:
            gain_loss: GainLoss = cast(GainLoss, entry)
            if not gain_loss.from_lot:
                continue
            if gain_loss.from_lot != current_from_lot:
                if current_from_lot:
                    border_style = self.__get_border_style(current_from_lot.timestamp.year, year)
                    year = border_style.year
                    border_suffix = border_style.border_suffix
                    self._fill_cell(sheet, in_transaction_index, 0, 1, data_style="percent", visual_style="from_lot" + border_suffix)
                    in_transaction_index += 1
                    current_from_lot_percentage = _ZERO
                current_from_lot = gain_loss.from_lot
            current_from_lot_percentage += gain_loss.from_lot_fraction_percentage
        if current_from_lot:
            border_style = self.__get_border_style(current_from_lot.timestamp.year, year)
            year = border_style.year
            border_suffix = border_style.border_suffix
        self._fill_cell(sheet, in_transaction_index, 0, current_from_lot_percentage, data_style="percent", visual_style="from_lot" + border_suffix)

        return row_index

    def __generate_out_table(self, sheet: Any, out_transaction_set: TransactionSet, row_index: int) -> int:
        row_index = self._fill_header("Out-Flow Detail", _OUT_HEADER_NAMES_ROW_1, _OUT_HEADER_NAMES_ROW_2, sheet, row_index, 1)

        entry: AbstractEntry
        crypto_out_running_sum: RP2Decimal = _ZERO
        crypto_fee_running_sum: RP2Decimal = _ZERO
        year: int = 0
        for entry in out_transaction_set:
            transaction: OutTransaction = cast(OutTransaction, entry)
            visual_style: str
            highlighted_style: str
            transaction_visual_style: _TransactionVisualStyle = self.__get_transaction_visual_style(transaction, year)
            year = transaction_visual_style.year
            visual_style = transaction_visual_style.visual_style
            highlighted_style = transaction_visual_style.highlighted_style
            self._fill_cell(sheet, row_index, 0, "", visual_style="transparent")
            self._fill_cell(sheet, row_index, 1, transaction.timestamp, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 2, transaction.asset, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 3, transaction.exchange, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 4, transaction.holder, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 5, transaction.transaction_type.value.upper(), visual_style=visual_style)
            self._fill_cell(sheet, row_index, 6, transaction.spot_price, visual_style=visual_style, data_style="usd")
            self._fill_cell(sheet, row_index, 7, transaction.crypto_out_no_fee, visual_style=visual_style, data_style="crypto")
            self._fill_cell(sheet, row_index, 8, transaction.crypto_fee, visual_style=visual_style, data_style="crypto")
            crypto_out_running_sum += transaction.crypto_out_no_fee
            crypto_fee_running_sum += transaction.crypto_fee
            self._fill_cell(sheet, row_index, 9, crypto_out_running_sum, data_style="crypto", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 10, crypto_fee_running_sum, data_style="crypto", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 11, transaction.crypto_out_no_fee * transaction.spot_price, visual_style=highlighted_style, data_style="usd")
            self._fill_cell(sheet, row_index, 12, transaction.crypto_fee * transaction.spot_price, visual_style=highlighted_style, data_style="usd")
            self._fill_cell(sheet, row_index, 13, "YES" if transaction.is_taxable() else "NO", data_style="usd", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 14, transaction.notes, visual_style="transparent")
            row_index += 1

        return row_index

    def __generate_intra_table(self, sheet: Any, intra_transaction_set: TransactionSet, row_index: int) -> int:
        row_index = self._fill_header("Intra-Flow Detail", _INTRA_HEADER_NAMES_ROW_1, _INTRA_HEADER_NAMES_ROW_2, sheet, row_index, 1)

        entry: AbstractEntry
        crypto_fee_running_sum: RP2Decimal = _ZERO
        year: int = 0
        for entry in intra_transaction_set:
            transaction: IntraTransaction = cast(IntraTransaction, entry)
            visual_style: str
            highlighted_style: str
            transaction_visual_style: _TransactionVisualStyle = self.__get_transaction_visual_style(transaction, year)
            year = transaction_visual_style.year
            visual_style = transaction_visual_style.visual_style
            highlighted_style = transaction_visual_style.highlighted_style
            self._fill_cell(sheet, row_index, 0, "", visual_style="transparent")
            self._fill_cell(sheet, row_index, 1, transaction.timestamp, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 2, transaction.asset, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 3, transaction.from_exchange, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 4, transaction.from_holder, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 5, transaction.to_exchange, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 6, transaction.to_holder, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 7, transaction.spot_price, visual_style=visual_style, data_style="usd")
            self._fill_cell(sheet, row_index, 8, transaction.crypto_sent, visual_style=visual_style, data_style="crypto")
            self._fill_cell(sheet, row_index, 9, transaction.crypto_received, visual_style=visual_style, data_style="crypto")
            self._fill_cell(sheet, row_index, 10, transaction.crypto_fee, visual_style=visual_style, data_style="crypto")
            crypto_fee_running_sum += transaction.crypto_fee
            self._fill_cell(sheet, row_index, 11, crypto_fee_running_sum, data_style="crypto", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 12, transaction.usd_fee, visual_style=highlighted_style, data_style="usd")
            self._fill_cell(sheet, row_index, 13, "YES" if transaction.is_taxable() else "NO", data_style="usd", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 14, transaction.notes, visual_style="transparent")
            row_index += 1

        return row_index

    def __generate_gain_loss_summary(self, sheet: Any, yearly_gain_loss_list: List[YearlyGainLoss], row_index: int) -> int:
        row_index = self._fill_header("Gain / Loss Summary", _GAIN_LOSS_SUMMARY_HEADER_NAMES_ROW_1, _GAIN_LOSS_SUMMARY_HEADER_NAMES_ROW_2, sheet, row_index, 0)

        year: int = 0
        for yearly_gain_loss in yearly_gain_loss_list:
            border_suffix: str = ""
            capital_gains_type: str = "LONG" if yearly_gain_loss.is_long_term_capital_gains else "SHORT"
            border_style: _BorderStyle = self.__get_border_style(yearly_gain_loss.year, year)
            year = border_style.year
            border_suffix = border_style.border_suffix
            self._fill_cell(sheet, row_index, 0, yearly_gain_loss.year, visual_style="transparent" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 1, yearly_gain_loss.asset, visual_style="transparent" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 2, yearly_gain_loss.usd_gain_loss, visual_style="bold" + border_suffix, data_style="usd")
            self._fill_cell(sheet, row_index, 3, capital_gains_type, visual_style="bold" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 4, yearly_gain_loss.transaction_type.value.upper(), visual_style="bold" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 5, yearly_gain_loss.crypto_amount, visual_style="transparent" + border_suffix, data_style="crypto")
            self._fill_cell(sheet, row_index, 6, yearly_gain_loss.usd_amount, visual_style="taxable_event" + border_suffix, data_style="usd")
            self._fill_cell(sheet, row_index, 7, yearly_gain_loss.usd_cost_basis, visual_style="from_lot" + border_suffix, data_style="usd")
            row_index += 1

        return row_index

    def __generate_account_balances(self, sheet: Any, balance_set: BalanceSet, row_index: int) -> int:
        row_index = self._fill_header("Account Balances", _BALANCE_HEADER_NAMES_ROW_1, _BALANCE_HEADER_NAMES_ROW_2, sheet, row_index, 0)

        totals: Dict[str, RP2Decimal] = {}
        value: RP2Decimal
        for balance in balance_set:
            self._fill_cell(sheet, row_index, 0, balance.exchange, visual_style="bold", data_style="default")
            self._fill_cell(sheet, row_index, 1, balance.holder, visual_style="bold", data_style="default")
            self._fill_cell(sheet, row_index, 2, balance.asset)
            self._fill_cell(sheet, row_index, 3, balance.acquired_balance, data_style="crypto")
            self._fill_cell(sheet, row_index, 4, balance.sent_balance, data_style="crypto")
            self._fill_cell(sheet, row_index, 5, balance.received_balance, data_style="crypto")
            self._fill_cell(sheet, row_index, 6, balance.final_balance, visual_style="bold", data_style="crypto")
            value = totals.setdefault(balance.holder, _ZERO)
            value += balance.final_balance
            totals[balance.holder] = value
            row_index += 1

        holder: str
        border_drawn: bool = False
        for (holder, value) in sorted(totals.items()):
            border_suffix: str = ""
            if not border_drawn:
                border_suffix = "_border"
                border_drawn = True
            self._fill_cell(sheet, row_index, 0, "Total", visual_style="bold" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 1, holder, visual_style="bold" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 2, "", visual_style="transparent" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 3, "", visual_style="transparent" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 4, "", visual_style="transparent" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 5, "", visual_style="transparent" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 6, value, visual_style="bold" + border_suffix, data_style="crypto")
            row_index += 1

        return row_index

    def __generate_average_price_per_unit(self, sheet: Any, asset: str, price_per_unit: RP2Decimal, row_index: int) -> int:
        self._fill_cell(sheet, row_index, 0, "Average Price", visual_style="title")
        self._fill_cell(sheet, row_index + 1, 0, "Average Price", visual_style="header")
        self._fill_cell(sheet, row_index + 2, 0, f"Paid Per 1 {asset}", visual_style="header")
        self._fill_cell(sheet, row_index + 3, 0, price_per_unit, visual_style="transparent", data_style="usd")

        return row_index + 4

    def __generate_gain_loss_detail(self, sheet: Any, asset: str, gain_loss_set: GainLossSet, row_index: int) -> int:

        row_index = self._fill_header("Gain / Loss Detail", _GAIN_LOSS_DETAIL_HEADER_NAMES_ROW_1, _GAIN_LOSS_DETAIL_HEADER_NAMES_ROW_2, sheet, row_index, 0)

        taxable_event_style_modifier: str = ""
        from_lot_style_modifier: str = ""
        crypto_amount_running_sum: RP2Decimal = _ZERO
        year: int = 0
        border_style: _BorderStyle
        for entry in gain_loss_set:
            gain_loss: GainLoss = cast(GainLoss, entry)
            border_suffix: str = ""
            border_style = self.__get_border_style(gain_loss.taxable_event.timestamp.year, year)
            year = border_style.year
            border_suffix = border_style.border_suffix
            transparent_style: str = f"transparent{border_suffix}"
            taxable_event_style: str = f"taxable_event{taxable_event_style_modifier}{border_suffix}"
            highlighted_style: str = f"highlighted{border_suffix}"
            current_taxable_event_fraction: int = gain_loss_set.get_taxable_event_fraction(gain_loss) + 1
            total_taxable_event_fractions: int = gain_loss_set.get_taxable_event_number_of_fractions(gain_loss.taxable_event)
            transaction_type: str = (
                f"{self._get_table_type_from_transaction(gain_loss.taxable_event)} / " f"{gain_loss.taxable_event.transaction_type.value.upper()}"
            )
            taxable_event_note: str = (
                f"{current_taxable_event_fraction}/"
                f"{total_taxable_event_fractions}: "
                f"{gain_loss.crypto_amount:.8f} of "
                f"{gain_loss.taxable_event.crypto_balance_change:.8f} "
                f"{asset}"
            )
            self._fill_cell(sheet, row_index, 0, gain_loss.crypto_amount, visual_style=transparent_style, data_style="crypto")
            self._fill_cell(sheet, row_index, 1, gain_loss.asset, visual_style=transparent_style)
            crypto_amount_running_sum += gain_loss.crypto_amount
            self._fill_cell(sheet, row_index, 2, crypto_amount_running_sum, visual_style=transparent_style, data_style="crypto")
            self._fill_cell(sheet, row_index, 3, gain_loss.usd_gain, visual_style=transparent_style, data_style="usd")
            self._fill_cell(sheet, row_index, 4, "LONG" if gain_loss.is_long_term_capital_gains() else "SHORT", visual_style=transparent_style)
            self._fill_cell(sheet, row_index, 5, gain_loss.taxable_event.timestamp, visual_style=taxable_event_style)
            self._fill_cell(sheet, row_index, 6, transaction_type, visual_style=taxable_event_style)
            self._fill_cell(sheet, row_index, 7, gain_loss.taxable_event_fraction_percentage, visual_style=taxable_event_style, data_style="percent")
            self._fill_cell(sheet, row_index, 8, gain_loss.taxable_event_usd_amount_with_fee_fraction, visual_style=highlighted_style, data_style="usd")
            self._fill_cell(sheet, row_index, 9, gain_loss.taxable_event.spot_price, visual_style=taxable_event_style, data_style="usd")
            self._fill_cell(sheet, row_index, 10, taxable_event_note, visual_style=f"taxable_event_note{border_suffix}")
            if current_taxable_event_fraction == total_taxable_event_fractions:
                # Last fraction: change color
                taxable_event_style_modifier = "" if taxable_event_style_modifier == "_alt" else "_alt"

            from_lot_style: str = f"from_lot{from_lot_style_modifier}{border_suffix}"
            if gain_loss.from_lot:
                current_from_lot_fraction: int = gain_loss_set.get_from_lot_fraction(gain_loss) + 1
                total_from_lot_fractions: int = gain_loss_set.get_from_lot_number_of_fractions(gain_loss.from_lot)
                from_lot_note: str = (
                    f"{current_from_lot_fraction}/"
                    f"{total_from_lot_fractions}: "
                    f"{gain_loss.crypto_amount:.8f} of "
                    f"{gain_loss.from_lot.crypto_balance_change:.8f} "
                    f"{asset}"
                )
                self._fill_cell(sheet, row_index, 11, gain_loss.from_lot.timestamp, visual_style=from_lot_style)
                self._fill_cell(sheet, row_index, 12, gain_loss.from_lot_fraction_percentage, visual_style=from_lot_style, data_style="percent")
                self._fill_cell(sheet, row_index, 13, gain_loss.from_lot_usd_amount_with_fee_fraction, visual_style=from_lot_style, data_style="usd")
                usd_fee_fraction: RP2Decimal = gain_loss.from_lot.usd_fee * gain_loss.from_lot_fraction_percentage
                self._fill_cell(sheet, row_index, 14, usd_fee_fraction, visual_style=from_lot_style, data_style="usd")
                self._fill_cell(sheet, row_index, 15, gain_loss.usd_cost_basis, visual_style=highlighted_style, data_style="usd")
                self._fill_cell(sheet, row_index, 16, gain_loss.from_lot.spot_price, visual_style=from_lot_style, data_style="usd")
                self._fill_cell(sheet, row_index, 17, from_lot_note, visual_style=f"from_lot_note{border_suffix}")

                if current_from_lot_fraction == total_from_lot_fractions:
                    # Last fraction: change color
                    from_lot_style_modifier = "" if from_lot_style_modifier == "_alt" else "_alt"
            else:
                for i in range(11, 17):
                    self._fill_cell(sheet, row_index, i, "", visual_style=f"{from_lot_style}{border_suffix}")

            row_index += 1

        return row_index

    def __generate_yearly_gain_loss_summary(self, sheet: Any, asset: str, yearly_gain_loss_list: List[YearlyGainLoss], row_index: int) -> int:
        for gain_loss in yearly_gain_loss_list:
            visual_style: str = "transparent"
            capital_gains_type: str = "LONG" if gain_loss.is_long_term_capital_gains else "SHORT"
            self._fill_cell(sheet, row_index, 0, gain_loss.year, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 1, asset, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 2, gain_loss.usd_gain_loss, visual_style=visual_style, data_style="usd")
            self._fill_cell(sheet, row_index, 3, capital_gains_type, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 4, gain_loss.transaction_type.value.upper(), visual_style=visual_style)
            self._fill_cell(sheet, row_index, 5, gain_loss.crypto_amount, visual_style=visual_style, data_style="crypto")
            self._fill_cell(sheet, row_index, 6, gain_loss.usd_amount, visual_style=visual_style, data_style="usd")
            self._fill_cell(sheet, row_index, 7, gain_loss.usd_cost_basis, visual_style=visual_style, data_style="usd")
            row_index += 1

        return row_index


def main() -> None:
    pass


if __name__ == "__main__":
    main()



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from dataclasses import dataclass

from rp2.configuration import Configuration
from rp2.entry_types import EntrySetType
from rp2.rp2_error import RP2TypeError
from rp2.transaction_set import TransactionSet


@dataclass(frozen=True, eq=True)
class InputData:
    asset: str
    in_transaction_set: TransactionSet
    out_transaction_set: TransactionSet
    intra_transaction_set: TransactionSet

    @classmethod
    def type_check(cls, name: str, instance: "InputData") -> "InputData":
        Configuration.type_check_parameter_name(name)
        if not isinstance(instance, cls):
            raise RP2TypeError(f"Parameter '{name}' is not of type {cls.__name__}: {instance}")
        return instance

    def __post_init__(self) -> None:
        Configuration.type_check_string("asset", self.asset)
        TransactionSet.type_check("in_transaction_set", self.in_transaction_set, EntrySetType.IN, self.asset, False)
        TransactionSet.type_check("out_transaction_set", self.out_transaction_set, EntrySetType.OUT, self.asset, True)
        TransactionSet.type_check("intra_transaction_set", self.intra_transaction_set, EntrySetType.INTRA, self.asset, True)
